/* {{{
 *
 * $Id: dv1394_video.cxx,v 1.8 2005/04/03 12:20:00 chutz Exp $
 * vim:sts=2:sw=2:fdm=marker:
 *
 * An attempt to provide a MiniDV driver through the dv1394 kernel module
 *
 * Requirements:
 * - dv1394 kernel module loaded and running
 * - libdv (tested with 0.102)
 *
 * Author: Georgi Georgiev <chutz@gg3.net>
 *
 */

#pragma implementation "dv1394_video.h"

#include "dv1394_video.h"

#define BUF_FRAMES 1
#define PAL_FRAME_SIZE 150000
#define NTSC_FRAME_SIZE 120000
#define MAX_FRAME_SIZE PAL_FRAME_SIZE

PCREATE_VIDINPUT_PLUGIN(DV, PVideoInputDv1394Device);

static PMutex mutex;
static PDictionary<PString, PString> *dico;
static BYTE rawFrameBuffer[BUF_FRAMES][MAX_FRAME_SIZE];
static int frameToWrite;
static int frameToRead;
static enum videoFormatType {
  AutoDetect = 0,
  formatPAL  = 1,
  formatNTSC = 2
} videoFormat;

#if DEBUG >= 2
static int frameDiff;
static int frameTotal;
#endif

/* }}} */
// class InputDVGrabber/*{{{*/
//
InputDVGrabber::InputDVGrabber() : PThread(1000, NoAutoDeleteThread) {/*{{{*/
#if DEBUG >= 1
  ::cerr << "Grabber: Init" << ::endl;
#endif
  grabberExitFlag = FALSE;
  fd = -1;
}
/*}}}*/
void InputDVGrabber::Main() {/*{{{*/
#if DEBUG >= 1
  ::cerr << "Grabber: Start" << ::endl;
#endif
  while (1) {
    grabberMutex.Wait();
    if (grabberExitFlag == TRUE) {
      grabberMutex.Signal();
      break;
    }
    grabberMutex.Signal();
    ReadAnotherFrame();
  }
#if DEBUG >= 1
  ::cerr << "Grabber: Exit" << ::endl;
#endif
}
/*}}}*/
void InputDVGrabber::ReadAnotherFrame() {/*{{{*/
  if (fd < 0) {
    Sleep(10);
    return;
  }

  BYTE buffer[MAX_FRAME_SIZE];
  BYTE * buffer_end = buffer;
  // PAL or NTSC?
  buffer_end += ::read(fd, buffer, 4);
  int thisFrameSize;
  mutex.Wait();
  switch (videoFormat) {
    case formatPAL:
      mutex.Signal();
#if DEBUG >= 2
      ::cerr << "User-specified PAL format" << ::endl;
#endif
      thisFrameSize = PAL_FRAME_SIZE;
      break;
    case formatNTSC:
      mutex.Signal();
#if DEBUG >= 2
      ::cerr << "User-specified NTSC format" << ::endl;
#endif
      thisFrameSize = NTSC_FRAME_SIZE;
      break;
    case AutoDetect:
    default:
      mutex.Signal();
      if (buffer[2] & 0x80) {
	thisFrameSize = PAL_FRAME_SIZE;
#if DEBUG >= 2
	::cerr << "Detected PAL format" << ::endl;
#endif
      } else {
	thisFrameSize = NTSC_FRAME_SIZE;
#if DEBUG >= 2
	::cerr << "Detected NTSC format" << ::endl;
#endif
      }
  }
  while (buffer_end - buffer < thisFrameSize) {
    buffer_end += ::read(fd, buffer_end, thisFrameSize - (buffer_end - buffer));
  }
  mutex.Wait();
  memcpy(rawFrameBuffer[frameToWrite], buffer, thisFrameSize);
#if ! BUF_FRAMES > 1
  frameToRead = 0;
#else
  if (frameToRead < 0)             { frameToRead = 0; }
  frameToWrite++;
#if DEBUG >= 2
  frameDiff++;
  frameTotal++;
  PTime t;
  ::cerr << "WRITING    "
    << t.GetHour() << ":" << t.GetMinute() << ":" << t.GetSecond() << "." << t.GetMicrosecond() << "\t"
    << "DIFF " << frameDiff << "/" << frameTotal << "\t" << (double)frameDiff/(double)frameTotal*100.0 << "%" << endl;
#endif
  if (frameToWrite >= BUF_FRAMES)  { frameToWrite = 0; }
  if (frameToRead == frameToWrite) { frameToRead++; }
  if (frameToRead >= BUF_FRAMES)   { frameToRead = 0; }
#endif
  mutex.Signal();
};
/*}}}*/
void InputDVGrabber::Stop() {/*{{{*/
#if DEBUG >= 1
  ::cerr << "Grabber: Stop" << ::endl;
#endif
  grabberMutex.Wait();
  grabberExitFlag = TRUE;
  grabberMutex.Signal();
}
/*}}}*//*}}}*/
// class PVideoInputDV1394/*{{{*/
///////////////////////////////////////////////////////////////////////////////

PVideoInputDv1394Device::PVideoInputDv1394Device()/*{{{*/
{
  fd = -1;
  is_capturing = FALSE;
  dv_decoder = NULL;
  grabber    = NULL;
}
/*}}}*/
PVideoInputDv1394Device::~PVideoInputDv1394Device()/*{{{*/
{
  Close();
}
/*}}}*/
BOOL PVideoInputDv1394Device::Open(const PString & ufName, BOOL startImmediate)/*{{{*/
{
  PTRACE(3, "trying to open " << ufName);
  PString * devName;

  if (IsOpen())
    Close();

  mutex.Wait();
  if(dico != NULL)
  {
    devName = dico->GetAt(ufName);
  }
  else
  {
    return FALSE;
  }
  mutex.Signal();
  frameWidth = CIFWidth;
  frameHeight = CIFHeight;
  colourFormat = "RGB24F";
  desiredFrameHeight = CIFHeight;
  desiredFrameWidth = CIFWidth;
  desiredColourFormat = "RGB24F";
  
  fd = ::open(*devName, O_RDONLY);
  if (fd == -1) {
    PTRACE(3, "Open() failed");
    return FALSE;
  }

  if (!SetChannel(channelNumber)
      || !SetVideoFormat(videoFormat)) {
    PTRACE(3, "SetChannel() or SetVideoFormat() failed");
    Close();
    return FALSE;
  }
  
  if (startImmediate && !Start()) {
    Close();
    return FALSE;
  }
  
  PTRACE(3, "Successfully opened DV1394");
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::IsOpen() /*{{{*/
{
  return fd != -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::Close()/*{{{*/
{
  PTRACE(3, "Close()");
  if (IsOpen()) {
    if (IsCapturing())
      Stop();
    if(::close(fd) == 0) {
      fd = -1;
      return TRUE;
    } else {
      return FALSE;
    }
  }
  return FALSE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::Start()/*{{{*/
{
  if (!IsOpen()) return FALSE;
  if (IsCapturing()) return TRUE;
  
  mutex.Wait();
  frameToWrite = 0;
#if BUF_FRAMES > 1
  frameToRead = -1;
  frameDiff   = 0;
  frameTotal  = 0;
#else
  frameToRead = 0;
#endif
  grabber = new InputDVGrabber();
  grabber->fd = fd;
  grabber->Resume();
  mutex.Signal();

  is_capturing = TRUE;
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::Stop()/*{{{*/
{
#if DEBUG >= 1
  ::cerr << "DV1394: Stop" << ::endl;
#endif
  if (IsCapturing()) {
    grabber->Stop();
    grabber->WaitForTermination();
    delete grabber;
    grabber = NULL;
    is_capturing = FALSE;
    return TRUE;
  }
  else
    return FALSE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::IsCapturing()/*{{{*/
{
  return is_capturing;
}
/*}}}*/
PString	    CheckInputDevice(PString dev) { /*{{{*/
    PString *devname = new PString(dev);
    PString ufname = *devname;
    struct stat dummy;
    if (::stat(*devname, &dummy) != 0) return "";
    mutex.Wait();
    if(dico == NULL) dico = new PDictionary<PString, PString>;
    dico->SetAt(ufname, devname);
    mutex.Signal();
    return ufname;
}
/*}}}*/
PStringList PVideoInputDv1394Device::GetInputDeviceNames()/*{{{*/
{
  PStringList Result;
  
  /*
   * Looking for devices created by the dv1394 kernel module:
   * - devfs used to create these in /dev/..../NTSC/{in,out} or something
   * - udev-056 and linux-kernel 2.6.12-rc1 created a node at /dev/dv1394-0
   *
   * Make this thing work with my setup, apologize and support the rest later
   *
   * TODO: Borrow the ReadDirectory code from v4l2
   */

  PString ufname;
  for (int i=0; ; i++) if ((ufname = CheckInputDevice("/dev/dv1394/" + (PString) i)) != "") Result.AppendString(ufname); else break;
  for (int i=0; ; i++) if ((ufname = CheckInputDevice("/dev/dv1394-" + (PString) i)) != "") Result.AppendString(ufname); else break;
  for (int i=0; ; i++) if ((ufname = CheckInputDevice("/dev/ieee1394/dv/host" + (PString) i + "/NTSC/in")) != "") Result.AppendString(ufname); else break;
  for (int i=0; ; i++) if ((ufname = CheckInputDevice("/dev/ieee1394/dv/host" + (PString) i + "/PAL/in")) != "") Result.AppendString(ufname); else break;
  if ((ufname = CheckInputDevice("./symlinked.dv")) != "") Result.AppendString(ufname);
  return Result;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetVideoFormat(VideoFormat newFormat)/*{{{*/
{
  // FIXME: isn't it inherited from PVideoDevice anyway?
  if (!PVideoDevice::SetVideoFormat(newFormat)) {
    PTRACE(3,"PVideoDevice::SetVideoFormat failed");
    return FALSE;
  }
  else
    return TRUE;
}
/*}}}*/
int  PVideoInputDv1394Device::GetBrightness()/*{{{*/
{
  return -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetBrightness(unsigned newBrightness)/*{{{*/
{
  return FALSE;
}
/*}}}*/
int  PVideoInputDv1394Device::GetHue()/*{{{*/
{
  return -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetHue(unsigned newHue)/*{{{*/
{
  return FALSE;
}
/*}}}*/
int  PVideoInputDv1394Device::GetContrast()/*{{{*/
{
  return -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetContrast(unsigned newContrast)/*{{{*/
{
  return FALSE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetColour(unsigned newColour) /*{{{*/
{
  return -1;
}
/*}}}*/
int  PVideoInputDv1394Device::GetColour()/*{{{*/
{
  return -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetWhiteness(unsigned newWhiteness) /*{{{*/
{
  return FALSE;
}
/*}}}*/
int  PVideoInputDv1394Device::GetWhiteness()/*{{{*/
{
  return -1;
}
/*}}}*/
BOOL PVideoInputDv1394Device::GetParameters (int *whiteness, int *brightness,/*{{{*/
                                       int *colour, int *contrast, int *hue)
{
  *whiteness = -1;
  *brightness = -1;
  *colour = -1;
  *hue = -1;
  return FALSE;
}
/*}}}*/
int  PVideoInputDv1394Device::GetNumChannels() /*{{{*/
{
  int Result;
  mutex.Wait();
  if(dico != NULL)
    Result = dico->GetSize();
  else
    Result = 0;

  mutex.Signal();
  return Result;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetChannel(int newChannel)/*{{{*/
{
  if (PVideoDevice::SetChannel(newChannel) == FALSE)
    return FALSE;

  if(IsCapturing()) {
    Stop();
    Start();
  }
  
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetFrameRate(unsigned rate)/*{{{*/
{
  return PVideoDevice::SetFrameRate(rate);
}
/*}}}*/
BOOL PVideoInputDv1394Device::GetFrameSizeLimits(unsigned & minWidth,/*{{{*/
						  unsigned & minHeight,
						  unsigned & maxWidth,
						  unsigned & maxHeight) 
{
  minWidth = 160;
  maxWidth = 320;
  minHeight = 120;
  maxHeight = 240;
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::GetFrame(PBYTEArray & frame)/*{{{*/
{
  PINDEX returned;
  if (!GetFrameData(frame.GetPointer(GetMaxFrameBytes()), &returned))
    return FALSE;

  frame.SetSize(returned);
  return TRUE;
}
/*}}}*/
PINDEX PVideoInputDv1394Device::GetMaxFrameBytes()/*{{{*/
{
  
  if(converter == NULL)
    return frameBytes;
  
  PINDEX bytes = converter->GetMaxDstFrameBytes();
  if (bytes > frameBytes)
    return bytes;
  else
    return frameBytes;
  
}
/*}}}*/
BOOL PVideoInputDv1394Device::GetFrameDataNoDelay(BYTE * buffer,/*{{{*/
						   PINDEX * bytesReturned)
{
  if (!IsCapturing()) return FALSE;

  /*
  int read_frames = 0;
  BYTE * capture_buffer;
  BYTE * capture_buffer_end = capture_buffer;
  */
  mutex.Wait();
  if (frameToRead < 0) {
    mutex.Signal();
    return FALSE;
  }
  BYTE rawFrame[MAX_FRAME_SIZE];
  memcpy(rawFrame, rawFrameBuffer[frameToRead], MAX_FRAME_SIZE);
#if BUF_FRAMES > 1
  frameToRead++;
  if (frameToRead >= BUF_FRAMES) { frameToRead = 0; }
#if DEBUG >= 2
  PTime t;
  frameDiff--;
  ::cerr << "READING    "
    << t.GetHour() << ":" << t.GetMinute() << ":" << t.GetSecond() << "." << t.GetMicrosecond() << "\t" <<
    "DIFF " << frameDiff << "/" << frameTotal << "\t" << (double)frameDiff/(double)frameTotal*100.0 << "%" << endl;
#endif
#endif
  mutex.Signal();
  
  /*
  while(capture_buffer_end - capture_buffer < 120000) {
    read_frames = ::read(fd, capture_buffer_end, 120000 - (capture_buffer_end - capture_buffer));
    capture_buffer_end += read_frames;
  }
  */

  dv_decoder_t *dv;
  dv = dv_decoder_new(TRUE, FALSE, FALSE);
  /* possible quality values from dv_types.h:
   * DV_QUALITY_
   *            FASTEST : grayscale, thumbnail-size image
   *            AC_1
   *            AC_2
   *            DC
   *            COLOR
   *            AC_MASK
   *            BEST    : COLOR | AC_2
   */
  dv->quality = DV_QUALITY_BEST;
  if(dv_parse_header(dv, rawFrame) < 0) {
    PTRACE(3, "cannot parse dv frame header");
    return FALSE;
  }
  
  dv_color_space_t color_space;
  BYTE * pixels[3];
  int  pitches[3];
  
  pitches[0] = dv->width * 3;
  pitches[1] = pitches[2] = 0;
  
  pixels[0] = (uint8_t *)malloc(dv->width * dv->height * 3);
  pixels[1] = NULL;
  pixels[2] = NULL;
  color_space = e_dv_color_rgb;
  
  dv_decode_full_frame(dv, rawFrame, color_space, pixels, pitches);
  
  // FIXME: this is a manual resize ; the original author wondered about it
  float xRatio = dv->width / (float)frameWidth;
  float yRatio = dv->height/ (float)frameHeight;
  for(uint y = 0; y < frameHeight; y++)
    for (uint x = 0; x < frameWidth; x++) {
      uint16_t sourceX = (uint16_t) (x * xRatio);
      uint16_t sourceY = (uint16_t) (y * yRatio);
      memcpy (pixels[0]+3*(y*frameWidth+x),
	      pixels[0]+3*(sourceY*dv->width+sourceX),
	      3);
      
      // FIXME: BGR <-> RGB done by hand
      BYTE temp;
      int offset= (y*frameWidth+x)*3;
      temp = pixels[0][offset+0];
      pixels[0][offset+0] = pixels[0][offset+2];
      pixels[0][offset+2] = temp;
    }
  if (converter != NULL) {
    converter->Convert((const BYTE *)pixels[0], buffer, bytesReturned);
    if (pixels[0] != NULL)
      free(pixels[0]);
  }
  else {
    PTRACE(3, "Converter must exist. Something goes wrong.");
    return FALSE;
  }
  
  return TRUE;
  
}
/*}}}*/
BOOL PVideoInputDv1394Device::GetFrameData(BYTE * buffer,/*{{{*/
					    PINDEX * bytesReturned)
{
  int capturing_duration = 10000; // FIXME: what is it for?
  // notice: funny things happen when it is set too low!
  
  if(frameRate>0) {
    if(msBetweenFrames > capturing_duration)
      PThread::Current()->Sleep(msBetweenFrames - capturing_duration);
    PTime start;
    if(!GetFrameDataNoDelay(buffer, bytesReturned))
      return FALSE;
    PTime end;
    capturing_duration = (int)((end-start).GetMilliSeconds());
    return TRUE;
  }
  return GetFrameDataNoDelay(buffer, bytesReturned);
}
/*}}}*/
void PVideoInputDv1394Device::ClearMapping()/*{{{*/
{
  // do nothing...
}
/*}}}*/
BOOL PVideoInputDv1394Device::TestAllFormats()/*{{{*/
{
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetColourFormat(const PString & newFormat)/*{{{*/
{
  if (newFormat != colourFormat)
    return FALSE;
  
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetFrameSize(unsigned width, unsigned height)/*{{{*/
{
  // FIXME: shouldn't it return FALSE when asked an unsupported frame size? 
  frameWidth = width;
  frameHeight = height;
  colourFormat = "RGB24F";
  frameBytes = PVideoDevice::CalculateFrameBytes(frameWidth,
						 frameHeight, colourFormat);
  
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetFrameSizeConverter(unsigned width,/*{{{*/
						     unsigned height,
						     BOOL bScaleNotCrop)
{

  SetFrameSize(width, height);

  if (converter != NULL) 
    delete converter;
  
  desiredFrameWidth = width;
  desiredFrameHeight = height;

  converter = PColourConverter::Create(colourFormat,
				       desiredColourFormat, width, height);
  if (converter == NULL) {
    PTRACE(3, "Failed to make a converter.");
    return FALSE;
  }

  if (converter->SetSrcFrameSize(width,height) == FALSE) {
    PTRACE(3, "Failed to set source frame size of a converter.");
    return FALSE;
  }
  
  if (converter->SetDstFrameSize(desiredFrameWidth,
				 desiredFrameHeight, FALSE) == FALSE) {
    PTRACE(3, "Failed to set destination frame size (+scaling) of a converter.");
    return FALSE;
  }
  
  return TRUE;
}
/*}}}*/
BOOL PVideoInputDv1394Device::SetColourFormatConverter(const PString &colourFmt)/*{{{*/
{
  desiredColourFormat = colourFmt;
  return SetFrameSizeConverter(desiredFrameWidth, desiredFrameHeight, FALSE);
}
/*}}}*//*}}}*/
// End Of File ///////////////////////////////////////////////////////////////
