/*
 * $Id: dv1394_video.h,v 1.5 2005/04/03 10:31:44 chutz Exp $
 * vim:sts=2:sw=2:
 *
 * Author: Georgi Georgiev <chutz@gg3.net>
 *
 */


#ifndef _PVIDEOIODV1394

#define _PVIDEOIODV1394

#ifdef __GNUC__
#pragma interface
#endif

#include <sys/utsname.h>
#include <libdv/dv.h>
#include <sys/stat.h>

#include <ptlib.h>
#include <ptlib/videoio.h>
#include <ptlib/vconvert.h>
#include <ptlib/file.h>
#if !P_USE_INLINES
#include <ptlib/contain.inl>
#endif


/*
 * The dv1394 module creates a device that provides DV data. However, reading
 * *on-request* is no good, since frames tend to get lost because it is like this:
 *   - the reader reads a frame (frame N)
 *   - the reader returns that frame, meanwhile frame N+1 is already started on
 *   the device, but nobody reads it.
 *   - next time the reader tries to read a frame, N+1 is missed, so it waits
 *   for frame N+2
 *
 *   At least this is what the problem was with the old 1394AVC implementation
 *
 *   Trying to avoid this problem by introducing a thread that reads continuously.
 */

class InputDVGrabber : public PThread
{
  PCLASSINFO(InputDVGrabber, PThread);

  public:
  InputDVGrabber();
  void	Main();
  void	ReadAnotherFrame();
  void	Stop();
  int		fd;

  private:
  PMutex	grabberMutex;
  BOOL	grabberExitFlag;
};

class PVideoInputDv1394Device : public PVideoInputDevice
{
  PCLASSINFO(PVideoInputDv1394Device, PVideoInputDevice);
  public:
  PVideoInputDv1394Device();
  ~PVideoInputDv1394Device();
  BOOL Open(
      const PString & deviceName,   /// Device name to open
      BOOL startImmediate = TRUE    /// Immediately start device
      );

  BOOL IsOpen();
  BOOL Close();
  BOOL Start();
  BOOL Stop();
  BOOL IsCapturing();
  static PStringList GetInputDeviceNames();
  PStringList GetDeviceNames() const
  { return GetInputDeviceNames(); }

  PINDEX GetMaxFrameBytes();
  BOOL GetFrame( PBYTEArray & frame);

  /**Grab a frame, after a delay as specified by the frame rate.
  */
  BOOL GetFrameData(
      BYTE * buffer,                 /// Buffer to receive frame
      PINDEX * bytesReturned = NULL  /// OPtional bytes returned.
      );

  /**Grab a frame. Do not delay according to the current frame rate parameter.
  */
  BOOL GetFrameDataNoDelay(
      BYTE * buffer,                 /// Buffer to receive frame
      PINDEX * bytesReturned = NULL  /// OPtional bytes returned.
      );


  int GetBrightness();
  BOOL SetBrightness(unsigned newBrightness);
  int GetWhiteness();
  BOOL SetWhiteness(unsigned newWhiteness);
  int GetColour();
  BOOL SetColour(unsigned newColour);
  int GetContrast();
  BOOL SetContrast(unsigned newContrast);
  int GetHue();
  BOOL SetHue(unsigned newHue);
  BOOL GetParameters (int *whiteness, int *brightness, int *colour, int *contrast, int *hue);
  BOOL GetFrameSizeLimits(
      unsigned & minWidth,   /// Variable to receive minimum width
      unsigned & minHeight,  /// Variable to receive minimum height
      unsigned & maxWidth,   /// Variable to receive maximum width
      unsigned & maxHeight   /// Variable to receive maximum height
      ) ;

  void ClearMapping();

  int GetNumChannels();
  BOOL SetChannel( int channelNumber);
  BOOL SetFrameRate( unsigned rate );
  BOOL SetVideoFormat( VideoFormat videoFormat);
  BOOL SetFrameSize(
      unsigned width,   /// New width of frame
      unsigned height   /// New height of frame
      );
  BOOL SetColourFormat(
      const PString & colourFormat   // New colour format for device.
      );

  BOOL TestAllFormats();

  /**Set the frame size to be used, trying converters if available.

    If the device does not support the size, a set of alternate resolutions
    are attempted.  A converter is setup if possible.
    */
  BOOL SetFrameSizeConverter(
      unsigned width,        /// New width of frame
      unsigned height,       /// New height of frame
      BOOL     bScaleNotCrop /// Scale or crop/pad preference
      );

  /**Set the colour format to be used, trying converters if available.

    This function will set the colour format on the device to one that
    is compatible with a registered converter, and install that converter
    so that the correct format is used.
    */
  BOOL SetColourFormatConverter(
      const PString & colourFormat // New colour format for device.
      );

  protected:

  BOOL		is_capturing;
  dv_decoder_t *	dv_decoder;
  int		fd;
  PString 	desiredColourFormat;
  PINDEX		frameBytes;
  unsigned	desiredFrameWidth;
  unsigned	desiredFrameHeight;
  InputDVGrabber	* grabber;
  BOOL SetupHandle();
};

#endif


// End Of File ///////////////////////////////////////////////////////////////
